#include "mat4x4.h"
///#include <GL/gl.h>
#include "quaternion.h"

NACB_BEGIN

Mat4x4::Mat4x4(const Vec3d & p,const Quaternion & q)
{
  Vec3d r0;
  Vec3d r1;
  Vec3d r2;
  q.getMatrixRows(r0,r1,r2);
  
  for(int j=0;j<3;j++){
    operator()(0,j)=r0.data[j];
    operator()(1,j)=r1.data[j];
    operator()(2,j)=r2.data[j];
    
    operator()(0,3)=p.x;
    operator()(1,3)=p.y;
    operator()(2,3)=p.z;
    
    operator()(3,j)=0;
  }
  operator()(3,3)=1;
}

Mat4x4 Mat4x4::eucInv() const {
  //no assertion, just assume euclidean and invert
  Mat4x4 inv=transpose();
  inv(3,0)=0;
  inv(3,1)=0;
  inv(3,2)=0;
  inv(3,3)=1;
  
  Vec4d cc=inv*Vec4d(get(0,3),get(1,3),get(2,3),0);
  inv(0,3)=-cc.x;
  inv(1,3)=-cc.y;
  inv(2,3)=-cc.z;
  return inv;
}

Mat4x4 Mat4x4::inverse() const{
  Mat4x4 inv=Mat4x4::eye();
  Mat4x4 mat;
  mat.operator=(*this);

  
  for(int j=0;j<4;j++){
    int best=0;
    float best_val=0;
    for(int i=j;i<4;i++){
      if(fabs(mat(i,j))>best_val){
	best=i;
	best_val=fabs(mat(i,j));
      }
    }
    //swap the two rows
    if(j!=best){
      for(int jj=0;jj<4;jj++){
	double t=mat(j,jj);
	mat(j,jj)=mat(best,jj);
	mat(best,jj)=t;
	  
	t=inv(j,jj);
	inv(j,jj)=inv(best,jj);
	inv(best,jj)=t;
      }
    }
    //anihilate above and below
    for(int i=0;i<4;i++){
      if(j==i)continue;
      double sc=-mat(i,j)/mat(j,j);
	
      for(int k=0;k<4;k++){
	mat(i,k)+=sc*mat(j,k);
	inv(i,k)+=sc*inv(j,k);
      }
    }
    //make row one
    double sc=1.0/mat(j,j);
    for(int k=0;k<4;k++){
      mat(j,k)*=sc;
      inv(j,k)*=sc;
    }
  }
  return inv;
}

void Mat4x4::print(const char * name) const{
  printf("%s=[\n",name);
  for(int i=0;i<4;i++){
    for(int j=0;j<4;j++){
      printf("%6.5f ",data[4*i+j]); 
    }
    printf("\n");
  }
  printf("]\n");
}


Mat4x4 Mat4x4::angleAxis(const Vec3d & axis,double t){
  Mat4x4 m;
  m.to_eye();
  Mat4x4 ss;
  ss.zero();
  
  double st=sin(t);
  double ct=cos(t);
  
  ss(0,1)=-axis.z;
  ss(0,2)= axis.y;
  
  ss(1,0)= axis.z;
  ss(1,2)=-axis.x;
  
  ss(2,0)=-axis.y;
  ss(2,1)= axis.x;
  
  Mat4x4 ss2=ss;
  ss2.mult_right(ss);
  m=m+ss*st+(ss2)*(1-ct);
  return m;
}

std::ostream & operator<<(std::ostream & os,const Mat4x4 & mat){
  mat.print(os);
  return os;
}

NACB_END

#ifdef TEST_MAT4x4
#include <iostream>
using namespace nacb;
using namespace std;

int test_assert(bool val,const char * mesg){
  if(!val)
    cout << "Error: " << mesg << "\n";
  return val;
}

int main(int ac,char * av[]){
  double eps=1e-7;
  int succ=0;
  Mat4x4 e;
  e.to_eye();
  
  Mat4x4 mat=Mat4x4::trans(0.4,-0.1,1)*Mat4x4::rotx(-0.1)*Mat4x4::roty(0.2);
  Mat4x4 minv=Mat4x4::roty(-0.2)*Mat4x4::rotx(0.1)*Mat4x4::trans(-0.4,0.1,-1);
  
  succ+=test_assert((mat*mat.eucInv()-Mat4x4::eye()).len()<eps,"euc inv");
  succ+=test_assert((mat*minv-Mat4x4::eye()).len()<eps,"an inv");

  mat*=minv;
  succ+=test_assert((mat-Mat4x4::eye()).len()<eps,"operator*=");

  for(int i=0;i<100;i++){
    Mat4x4 m=Mat4x4::rand(0,i+1);
    succ+=test_assert((m*m.inverse()-Mat4x4::eye()).len()<eps,"inverse");
  }

  cout << "Successful tests: "<<succ<<"\n";
  
  return 0;
}
#endif
