#include "sparsematrix.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <algorithm>

#ifdef USE_UMFPACK
#include <ufsparse/umfpack.h>
#endif

extern "C" {
#include "clsqr1/lsqr.h"
}

//ARPACK function declarations
extern "C" void dsaupd_ (int & ido,char & bmat,
			 int & n,
			 char * which,
			 int & nev,
			 double & tol, 
			 double * resid, int & ncv,
			 double * v,
			 int & ldv,
			 int * iparam, 
			 int * ipntr, 
			 double * workd, 
			 double * workl,
			 int & lworkl, int & info);
//I originally thought rvec and select were chars, but it was crashing
//sporadically until I set it to be integers (I found some other declarations
//online that had integers as well)
extern "C" void dseupd_ (const int & rvec,const char & howmny,
			 int* select,double * d,double * z,int & ldz,
			 double & sigma,
			 char & bmat,
			 int & n,
			 char * which,
			 int & nev,
			 double & tol, 
			 double * resid, int & ncv,
			 double * v,
			 int & ldv,
			 int * iparam, 
			 int * ipntr, 
			 double * workd, 
			 double * workl,
			 int & lworkl, int & info);

NACB_BEGIN

static void lsqr_vec_prod(long mode,dvec * x,dvec * y,void * d){
  SparseMatrix * mat = reinterpret_cast<SparseMatrix *>(d);
  mat->lsqr_vec_prod_add(mode, x->elements, y->elements);
}


SparseMatrix::SparseMatrix(int m,int n,int nzmax){
  if(nzmax<=1)nzmax=1;
  this->m=m;
  this->n=n;
  this->nzmax=0;
  nnz=0;
  row=0;
  col=0;
  data=0;
  ensure(nzmax);
}

SparseMatrix::SparseMatrix(const Matrix & mat){
  int matnnz=0;
  
  for(int i=0;i<mat.m;i++){
    for(int j=0;j<mat.n;j++){
      if(mat.get(i,j)!=0){
	matnnz++;
      }
    }
  }
  this->m=mat.m;
  this->n=mat.n;
  this->nnz=0;
  this->nzmax=0;
  this->row=0;
  this->col=0;
  this->data=0;
  
  ensure(matnnz);
  
  this->nnz=matnnz;
  int ind=0;
  for(int j=0;j<mat.n;j++){
    col[j]=ind;
    for(int i=0;i<mat.m;i++){
      if(mat.get(i,j)!=0){
	row[ind]=i;
	data[ind]=mat.get(i,j);
	ind++;
      }
    }
  }
  col[n]=ind;
}

SparseMatrix::SparseMatrix(const SparseMatrix & sm){
  this->m=this->n=this->nzmax=this->nnz=0;
  this->row=this->col=0;
  this->data=0;
  operator=(sm);
}

void SparseMatrix::operator=(const SparseMatrix & sm){
  if(&sm==this)return;
  release();
  ensure(sm.nzmax);
  this->m=sm.m;
  this->n=sm.n;
  this->nnz=sm.nnz;
  memcpy(data,sm.data,sizeof(double)*nzmax);
  memcpy(data,sm.row,sizeof(int)*nzmax);
  memcpy(data,sm.col,sizeof(int)*(n+1));
}

void SparseMatrix::printMatlab(char * str){
  printf("%s=[\n",str);
  for(int i=0;i<m;i++){
    for(int j=0;j<n;j++){
      if(have(i,j))
	printf("%f ",get(i,j));
      else printf("%f ",0.0f);
    }
    printf("\n");
  }
  printf("];\n");
}

void SparseMatrix::ensure(int nzmax){
  if(this->nzmax>=nzmax)return;
  int * rowold=row;
  double * dataold=data;
  data=new double[nzmax];
  row=new int[nzmax];
  if(!col){
    col=new int[n+1];
    memset(col,0,sizeof(int)*(n+1));
  }
  if(rowold && dataold){
    memcpy(data,dataold,sizeof(double)*this->nzmax);   
    memcpy(row,rowold,sizeof(int)*this->nzmax);
  }
  else{
    memset(data,0,sizeof(double)*nzmax);
    memset(row,0,sizeof(int)*nzmax);
  }      
  this->nzmax=nzmax;
  delete [] rowold;
  delete [] dataold;
}

double & SparseMatrix::insert_no_check(int i,int j,const double & d){
  int end=col[j+1];
  int ind=col[j];
  for(ind=col[j];ind<end;ind++){
    if(row[ind]>i)break;
  }
  //need to insert at ind,
  // ind is either bigger than i or in next column
  if(nnz>=nzmax)ensure(std::max(nnz+8,(int)(((double)nnz)*1.2)));
  for(int finger=col[n];finger>ind;finger--){
    row[finger]=row[finger-1];
    data[finger]=data[finger-1];
  }
  row[ind]=i;
  data[ind]=d;
  //need to shift all col inds up
  for(j=j+1;j<=n;j++){
    col[j]++;
  }
  nnz++;
  return data[ind];
}
double & SparseMatrix::get(int i,int j){
  //check dimensions?
  int end=col[j+1];
  for(int ind=col[j];ind<end;ind++){
    if(row[ind]==i){
      return data[ind];
    }
    //assume sorted
    else if(row[ind]>i){
      //didn't find it, need to add it
      return insert_no_check(i,j);
    }
  }
  //comes here if there is nothing bigger than i the column, 
  return insert_no_check(i,j);
}

void SparseMatrix::lsqr_vec_prod_add(long mode,double * X,double * Y){
  if(!mode){
    for(int j=0;j<n;j++){
      int end=col[j+1];
      for(int ind=col[j];ind<end;ind++){
	int  i=row[ind];
	Y[i]+=data[ind]*X[j];
      }
    }
  }
  else{
    for(int j=0;j<n;j++){
      double sum=0;
      int end=col[j+1];
      for(int ind=col[j];ind<end;ind++){
	sum+=data[ind]*Y[row[ind]];
      }
      X[j]+=sum;
    }
  }
}


void SparseMatrix::matprod(int trans,double * X,double * Y){
  if(!trans){
    memset(Y,0,sizeof(double)*m);

    for(int j=0;j<n;j++){
      int end=col[j+1];
      for(int ind=col[j];ind<end;ind++){
	int  i=row[ind];
	Y[i]+=data[ind]*X[j];
      }
    }
  }
  else{
    for(int j=0;j<n;j++){
      double sum=0;
      int end=col[j+1];
      for(int ind=col[j];ind<end;ind++){
	sum+=data[ind]*X[row[ind]];
      }
      Y[j]=sum;
    }
  }
}


Matrix SparseMatrix::lsqr(Matrix & b,double xtol,int maxits){
  assert(m==b.m);

  lsqr_input * lin;
  lsqr_output * lout;
  lsqr_work * lwork;
  lsqr_func * lfunc;

  alloc_lsqr_mem( &lin, &lout, &lwork, &lfunc, m, n);
  lfunc->mat_vec_prod=lsqr_vec_prod;
  lin->num_rows=m;
  lin->num_cols=n;
  lin->damp_val=0;
  lin->rel_mat_err=1e-9;
  lin->rel_rhs_err=1e-9;
  lin->cond_lim=100000;
  lin->max_iter=4*(lin->num_rows+lin->num_cols+50);
  lin->lsqr_fp_out=0;//stdout;
  lin->rhs_vec=(dvec *) alloc_dvec(m);
  lin->sol_vec=(dvec *) alloc_dvec(n);

  memcpy(lin->rhs_vec->elements, b.data, sizeof(double)*m);
  memset(lin->sol_vec->elements, 0, sizeof(double)*n);

  ::lsqr(lin,lout,lwork,lfunc,this);

  Matrix x(n,1);
  memcpy(x.data, lout->sol_vec->elements, sizeof(double)*n);

  free_lsqr_mem(lin, lout, lwork, lfunc);

  return x;
}

Matrix SparseMatrix::linLeastSq(Matrix & b,double xtol,int maxits){
  Matrix Ax(b.m,1);
  Matrix As(b.m,1);
  Matrix bA(n,1);
  Matrix g(n,1);
  Matrix s(n,1);
  Matrix x(n,1);
  double gdotg=1;

  x.setAll(0);

  matprod(true, b.data, bA.data);

  printf("this doesnt work as well as I had initially thought\n");

  for(int its=0; its<maxits; its++){
    matprod(false, x.data, Ax.data);
    
    matprod(true, Ax.data, g.data);
    g-=bA;
    g*=2.0;

    if(its==0 || gdotg<1e-8){
      s=g*-1;
      gdotg = g.dot(g);
    } 
    else {
      double gdg = g.dot(g);
      s= s*(gdg/gdotg)-g;
      gdotg = gdg;
    }
    
    matprod(false, s.data, As.data);
    Matrix bAx = b-Ax;

    //printf("res: %d, %f\n",its, bAx.dot(bAx));
    double alpha = (As.dot(bAx)/std::max(1e-14,As.dot(As)));
    double slen = sqrt(s.dot(s));

    x+=s*alpha;
    //printf("moved %f\n",alpha*slen);
    if(alpha*slen < xtol)break;
  }
  return x;
}

Matrix SparseMatrix::gaussSeidel(Matrix & b,double xtol,int maxits){
  assert(m==n);
  
  Matrix x(n,1);
  int *  ninrow = new int[m];
  int ** colind = new int*[m];
  int ** dind = new int*[m];

  memset(ninrow, 0, sizeof(int)*m);
  x.setAll(0);

  for(int j=0; j<n; j++)
    for(int k=col[j]; k<col[j+1]; k++)
      ninrow[row[k]]++;

  for(int i=0; i<m; i++){
    colind[i] = new int[ninrow[i]];
    dind[i] = new int[ninrow[i]];
    ninrow[i] = 0;
  }
  
  int tot = 0;
  for(int j=0; j<n; j++)
    for(int k=col[j]; k<col[j+1]; k++){
      colind[row[k]][ninrow[row[k]]] = j;
      dind[row[k]][ninrow[row[k]]] = k;
      ninrow[row[k]]++;
      tot++;
    }

  printf("tot is %d/%d\n",tot,nnz);
    

  for(int its=0; its<maxits; its++){
    double dx = 0;
    for(int i=0; i<m; i++){
      double bd = b[i];
      double Aii = 1e-5;
      for(int k=0; k<ninrow[i]; k++){
	int j = colind[i][k];
	int di = dind[i][k];

	if(j!=i)
	  bd -= data[di]*x[j];
	else
	  Aii = data[di];
      }
      double diff; 
      bd/=Aii;
      diff = bd-x[i];
      diff*=diff;
      ///printf("%f %f %f\n",x[i],bd,Aii);

      x[i] = bd;

      

      dx+=diff;
    }
    if(sqrt(dx)<xtol)
      break;
  }

  for(int i=0; i<m; i++){
    delete [] colind[i];
    delete [] dind[i];
  }

  delete [] colind;
  delete [] ninrow;
  delete [] dind;

  return x;
}

Matrix SparseMatrix::svd(double & sing_val,int smallest){
  int ido=0;
  char bmat='I';
  char which[3]="LA";
  double bound2=0;
  if(smallest){
    double bound=0;
    svd(bound,0);
    printf("biggest is %f\n",bound);
    bound2=bound*bound*1.1;
  }
  int nev=1;
  double tol=1e-12;
  int ncv=std::min(4*nev,n);//?How to choose good
  Matrix resid(n,1);
  int mode=1;
  int iparam[11]={1,0,2000,1,0,0,mode,0,0,0};
  int ipntr[11]={0,0,0,0,0,0,0,0,0,0};
  Matrix workd(3*n,1);
  int lworkl=ncv*(ncv+8);
  Matrix workl(lworkl,1);
  int info=0;
  Matrix v(ncv,n);
  int ldv=n;
  
  Matrix tvec(m,1);
  int it=0;
  while(1){
    it++;
#ifdef USE_ARPACK
    dsaupd_ (ido,bmat,n,which,nev,
	     tol,resid.data,ncv,
	     v.data,ldv,iparam,
	     ipntr,
	     workd.data,workl.data,
	     lworkl,info);
#endif
    if(ido==1||ido==-1){
      if(mode==1){
	matprod(0,workd.data+ipntr[0]-1,tvec.data);
	matprod(1,tvec.data,workd.data+ipntr[1]-1);
	if(smallest){
	  double * pntr0=workd.data+ipntr[0]-1;
	  double * pntr1=workd.data+ipntr[1]-1;
	  for(int i=0;i<n;i++){
	    pntr1[i]=bound2*pntr0[i]-pntr1[i];
	    if(isnan(pntr1[i]) || isinf(pntr1[i])){
	      printf("its nan\n");
	    }
	  }
	}
      }
      else if(mode==3){
	//(A-sigma*B)*v=M;
	//v=(A-sigma*B)\M;
	printf("mode is 3\n");
      }
    }
    else if(ido==99)break;
    else printf("unhandled ido %d\n",ido);
    if(info)break;
  }

  int rvec=1;
  int * select=new int[n];
  Matrix d(n,1);//ncv is longer than necessary
  int ldz=n;
  double sigma=0;//not referencesd
  int ierr=0;
  
  memset(select,0,sizeof(int)*n);
#ifdef USE_ARPACK
  dseupd_(rvec,'A', select, d.data, v.data, ldv, sigma, 
	  bmat, n, which, nev, tol, resid.data, ncv, v.data, ldv, 
	  iparam, ipntr, workd.data, workl.data, lworkl, ierr);
#endif
  v=v.transpose();
  
  if(info)
    printf("info is %d,ido is %d\n",info,ido);
  if(ierr)
    printf("ierr is %d\n",ierr);

  delete [] select;
  printf("deleted select\n");
  if(smallest)
    sing_val=sqrt(bound2-d[0]);
  else
    sing_val=sqrt(d[0]);
  
  return v.getColumn(0);
}

Matrix SparseMatrix::umfsolve(SparseMatrix & A,Matrix & b){
  void * symb,*num;
  assert(A.m==A.n);

#ifdef USE_UMFPACK
  umfpack_di_symbolic(A.n,A.n,A.col,A.row,A.data,
		      &symb,(double *)0,(double*)0);
  umfpack_di_numeric(A.col,A.row,A.data,symb,
		     &num,(double *)0,(double*)0);
  umfpack_di_free_symbolic(&symb);
  
  Matrix x(A.m,1);
  umfpack_di_solve(UMFPACK_A,A.col,A.row,A.data,x.data,b.data,
		   num,(double*)0,(double*)0);
  umfpack_di_free_numeric(&num);
  return x;
#else
  static int warned = 0;
  if(!warned){
    fprintf(stderr,"WARNING: no support for umfpack, using lsqr for linear systems.\n");
    fprintf(stderr,"         This message will appear once.\n");
    warned = 1;
  }
  return A.lsqr(b);
#endif
}


NACB_END

#ifdef TEST_SPARSE_SVD
int main(int ac,char * av[])
{
  printf("testing the sparse matrix svd function\n");
  int nobs=18;
  int ncams=100;
  Matrix M(ncams*nobs,4*ncams+6);
  for(int i=0;i<ncams;i++){
    for(int j=0;j<nobs;j++){
      int offs=RAND_MAX/2;
      M(nobs*i+j,0)=(rand()%RAND_MAX)-offs;
      M(nobs*i+j,1)=(rand()%RAND_MAX)-offs;
      M(nobs*i+j,2)=(rand()%RAND_MAX)-offs;
      M(nobs*i+j,3)=(rand()%RAND_MAX)-offs;
      M(nobs*i+j,4)=(rand()%RAND_MAX)-offs;
      M(nobs*i+j,5)=(rand()%RAND_MAX)-offs;
      M(nobs*i+j,4*i+6)=(rand()%RAND_MAX)-offs;
      M(nobs*i+j,4*i+7)=(rand()%RAND_MAX)-offs;
      M(nobs*i+j,4*i+8)=(rand()%RAND_MAX)-offs;
      M(nobs*i+j,4*i+9)=(rand()%RAND_MAX)-offs;
    }
  }
  M*=0.0001;
  /*for(int i=0;i<M.m*M.n;i++)
    M[i]=(M[i]-0.5)*10000;
  for(int i=0;i<M.m*200;i++){
    M(rand()%M.m,rand()%M.n)=0;
    }*/
  printf("initing the sparse matrix\n");
  SparseMatrix sparseM(M);
  printf("got the sparse matrix\n");

  printf("doing full svd\n");
  Matrix U,S,V;
  M.Lsvd(U,S,V);
  printf("done full svd\n");
  
  Matrix Mv=V.getColumn(V.n-1);
  double Ms=S[MAX(S.m,S.n)-1];
  
  double SMs;
  printf("getting last\n");
  Matrix SMv=sparseM.svd_last(SMs);
  printf("biggest in M %f\n",S[0]);
  printf("last singular value full:%f, sparse:%f\n",Ms,SMs);
  //Mv.printMatlab("Mv");
  //SMv.printMatlab("SMv");
  Matrix diff=Mv-SMv;
  Matrix odiff=Mv+SMv;
  printf("difference %f, dot %f\n",
	 std::min(sqrt((diff.dot(diff))/Mv.m),sqrt((odiff.dot(odiff))/Mv.m)),
	 Mv.dot(SMv));
  printf("exiting %d (%dx%d)\n",sparseM.nnz,sparseM.m,sparseM.n);
  return 0;
}
#endif

#ifdef TEST_SPARSE_LSQ
#include <sys/time.h>
int main(int ac,char * av[]){
  Matrix A=Matrix::random(1000,1000);
  Matrix b=Matrix::random(1000,1);
  
  for(int i=0;i<A.m*A.n;i++)
    if(rand()%100<10)A.data[i]=0;
  
  printf("solving for x1...\n");

  struct timeval st,et,dt;
  gettimeofday(&st,0);
  Matrix x1=Matrix::LlinLeastSq(A,b);
  gettimeofday(&et,0);
  timersub(&et,&st,&dt);

  printf("...got x1 took %f\n",double(dt.tv_sec)+double(dt.tv_usec)/1e6);
  
  Matrix diff;
  SparseMatrix sA(A);
#ifdef USE_UMFPACK
  printf("solving for x2...\n");
  gettimeofday(&st,0);
  Matrix x2=SparseMatrix::umfsolve(sA,b);
  gettimeofday(&et,0);
  timersub(&et,&st,&dt);
  printf("...got x2 took %f\n",double(dt.tv_sec)+double(dt.tv_usec)/1e6);
  diff=x1-x2;
  
  printf("diff: %e (%e,%e)\n",sqrt(diff.dot(diff)),
	 sqrt(x1.dot(x1)),sqrt(x2.dot(x2)));  
#endif
  gettimeofday(&st,0);
  Matrix x3=sA.linLeastSq(b,1e-8,10000);
  gettimeofday(&et,0);
  timersub(&et,&st,&dt);
  printf("...got x3 took %f\n",double(dt.tv_sec)+double(dt.tv_usec)/1e6);
  
  diff = x1-x3;
  Matrix tmp = A*x1-b;
  double r1 = tmp.dot(tmp);
  tmp = A*x3-b;
  double r3 = tmp.dot(tmp);

  printf("diff: %e (%e,%e) residuals %e,%e\n",sqrt(diff.dot(diff)),
	 sqrt(x1.dot(x1)),sqrt(x3.dot(x3)),r1,r3);

  return 0;
}

#endif
