#include <math.h>
#include <iostream>
#include <sstream>

#include "nspace.h"

#ifndef NACB_VEC4_H
#define NACB_VEC4_H

NACB_BEGIN

template <class T>
class Vec4{
public:
#ifdef __PY_PLUS_PLUS
  T x,y,z,w;
#else
  union{
    struct{
      T x,y,z,w;
    };
    T data[4];
  };
#endif
  Vec4(T xin=0,T yin=0,T zin=0,T win=1):x(xin),y(yin),z(zin),w(win){}
  Vec4(const Vec4 & o):x(o.x),y(o.y),z(o.z),w(o.w){}
  template <class T2>
    Vec4(const Vec4<T2> & o):x(o.x),y(o.y),z(o.z),w(o.w){}
  
  inline void rotatex(double t){rotx(t);}
  inline void rotatey(double t){roty(t);}
  inline void rotatez(double t){rotz(t);}

  /** \brief Rotate the vector by t around x.
   */
  void rotx(double t);

  /** \brief Rotate the vector by t around y.
   */
  void roty(double t);

  /** \brief Rotate the vector by t around z
   */
  void rotz(double t);

  /** \brief Set vector to (d/dt)rx(t)*vec. 
   */
  void dx(double t);

  /** \brief Set vector to (d/dt)ry(t)*vec. 
   */
  void dy(double t);

  /** \brief Set vector to (d/dt)rz(t)*vec. 
   */
  void dz(double t);

  double len()const{
    return sqrt(x*x+y*y+z*z+w*w);
  }
  double normalize(){
    double l=len();
    operator/=(l);
    return l;
  }

  /** \brief Multiply by s.
   */
  Vec4 operator*(double s) const {
    return Vec4(x*s,y*s,z*s,w*s);
  }
  /** \brief Pointwise product.
   */
  Vec4 operator*(const Vec4 & o)const{
    return Vec4(x*o.x,y*o.y,z*o.z,w*o.w);
  }
  /** \brief Divide by s.
   */
  Vec4 operator/(double s) const {
    return Vec4(x/s,y/s,z/s,w/s);
  }
  /** \brief Cross product, assuming vectors.
   */
  Vec4 operator^(const Vec4 & o) const {
    return Vec4(y*o.z-z*o.y,
		 -x*o.z+z*o.x,
		 x*o.y-y*o.x,0);
  }
  inline Vec4 cross(const Vec4 & o) const{return operator^(o);}
  
  Vec4 operator+(const Vec4 & o) const {
    return Vec4(x+o.x,y+o.y,z+o.z,w+o.w);
  }
  Vec4 operator-(const Vec4 & o) const {
    return Vec4(x-o.x,y-o.y,z-o.z,w-o.w);
  }
  /** \brief Multiply by vector s (in place).
   */
  void operator*=(const Vec4 & s){
    x*=s.x;
    y*=s.y;
    z*=s.z;
    w*=s.w;
  }
  /** \brief Multiply by s (in place).
   */
  void operator*=(double s){
    x*=s;
    y*=s;
    z*=s;
    w*=s;
  }
  /** \brief Divide by s (in place).
   */
  void operator/=(double s){
    x/=s;
    y/=s;
    z/=s;
    w/=s;
  }
  /** \brief Add o to this (in place)
   */
  void operator+=(const Vec4 & o){
    x+=o.x;
    y+=o.y;
    z+=o.z;
    w+=o.w;
  }
  /** \brief Add o to this (in place)
   */
  void operator-=(const Vec4 & o){
    x-=o.x;
    y-=o.y;
    z-=o.z;
    w-=o.w;
  }
  void print(std::ostream & os) const {
    os << "[" <<x<< " " << y << " " << z << " " << w <<"]";
  }
  operator std::string () const{
    std::ostringstream os;
    print(os);
    return os.str();
  }
  template<class T2>
    operator Vec4<T2> () const{
    return Vec4<T2>((T2)x,(T2)y,(T2)z,(T2)w);
  }
  bool operator==(const Vec4 & o) const{
    return x==o.x && y==o.y && z==o.z && w==o.w;
  }

  T * getData(){
#ifdef __PY_PLUS_PLUS
    return &x;
#else
    return data;
#endif
  }

  T & operator[](int i){
    return getData()[i];
  }
#ifdef __PY_PLUS_PLUS
  void python_set(int i,T t){
    getData()[i]=t;
  }
#endif
};

template <class T>
Vec4<T> operator*(double s,const Vec4<T> & vec){
  return vec*s;
}

template <class T>
std::ostream & operator<<(std::ostream & os,const Vec4<T> & vec){
  vec.print(os);
  return os;
}

typedef Vec4<float>  Vec4f;
typedef Vec4<double> Vec4d;


#define VEC4_INSTANTIATIONS template class Vec4<double>;\
  template class Vec4<float>;

#ifdef VEC4_INSTANTIATE
VEC4_INSTANTIATIONS
#endif

NACB_END

#endif // NACB_VEC4_H
